/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2007 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Class
    quadraticFit

Description
    Quadratic fit interpolation scheme which applies an explicit correction to
    linear.

SourceFiles
    quadraticFit.C

\*---------------------------------------------------------------------------*/

#ifndef quadraticFit_H
#define quadraticFit_H

#include "linear.H"
#include "quadraticFitData.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
                           Class quadraticFit Declaration
\*---------------------------------------------------------------------------*/

template<class Type>
class quadraticFit
:
    public linear<Type>
{
    // Private Data
        const scalar centralWeight_;

    // Private Member Functions

        //- Disallow default bitwise copy construct
        quadraticFit(const quadraticFit&);

        //- Disallow default bitwise assignment
        void operator=(const quadraticFit&);


public:

    //- Runtime type information
    TypeName("quadraticFit");


    // Constructors

        //- Construct from mesh and Istream
        quadraticFit(const fvMesh& mesh, Istream& is)
        :
            linear<Type>(mesh),
            centralWeight_(readScalar(is))
        {}


        //- Construct from mesh, faceFlux and Istream
        quadraticFit
        (
            const fvMesh& mesh,
            const surfaceScalarField& faceFlux,
            Istream& is
        )
        :
            linear<Type>(mesh),
            centralWeight_(readScalar(is))
        {}


    // Member Functions

        //- Return true if this scheme uses an explicit correction
        virtual bool corrected() const
        {
            return true;
        }

        //- Return the explicit correction to the face-interpolate
        virtual tmp<GeometricField<Type, fvsPatchField, surfaceMesh> >
        correction
        (
            const GeometricField<Type, fvPatchField, volMesh>& vf
        ) const
        {
            const fvMesh& mesh = this->mesh();

            const quadraticFitData& cfd = quadraticFitData::New
            (
                mesh, 
                centralWeight_
            );

            const labelListList& stCells = cfd.stencilCells();
            const labelListList& stBounds = cfd.stencilBounds();
            const List<scalarList>& f = cfd.fit();

            tmp<GeometricField<Type, fvsPatchField, surfaceMesh> > tsfCorr
            (
                new GeometricField<Type, fvsPatchField, surfaceMesh>
                (
                    IOobject
                    (
                        vf.name(),
                        mesh.time().timeName(),
                        mesh
                    ),
                    mesh,
                    dimensioned<Type>
                    (
                        vf.name(),
                        vf.dimensions(),
                        pTraits<Type>::zero
                    )
                )
            );
            
            GeometricField<Type, fvsPatchField, surfaceMesh>& sf = tsfCorr();
            
            for(label faci = 0; faci < mesh.nInternalFaces(); faci++)
            {
                // Use the fit coefficients to interpolate from the values
                // of the field at each cell in stencil

                for(label ip = 0; ip < stCells[faci].size(); ip++)
                {
                    sf[faci] += f[faci][ip] * vf[stCells[faci][ip]];
                }
                for(label ib = 0; ib < stBounds[faci].size(); ib++)
                {
                    label ip = ib + stCells[faci].size();
                    label facei = stBounds[faci][ib];
                    label patchi = mesh.boundaryMesh().whichPatch(facei);
                    label patchFacei = facei 
                                     - mesh.boundaryMesh()[patchi].start();

                    sf[faci] += f[faci][ip]
                             * vf.boundaryField()[patchi][patchFacei];
                }
            }
            
            return tsfCorr;
        }
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
